@@ -19,8 +19,6 @@
 
 Mail::SpamAssassin::BayesStore::PgSQL - PostgreSQL Specific Bayesian Storage Module Implementation
 
-=head1 SYNOPSIS
-
 =head1 DESCRIPTION
 
 This module implements a PostgreSQL specific bayesian storage module.
@@ -30,21 +28,24 @@ which makes SQL calls involving the token column.  Since PostgreSQL uses BYTEA
 for the token column type you must make sure that the DBD driver does the proper
 quoting.  You can accomplish this by binding the token column to a specific type.
 
+In versions 8.3 and up, synchronous_commit is disabled for faster performance.
+Data is not guaranteed to be stored until (3 * wal_writer_delay) (default: 0.6s)
+after it is written. This tradeoff works for Bayes because we care about aggregate
+data and a small period of time can be lost without sacrificing accuracy.
+
 =cut
 
 package Mail::SpamAssassin::BayesStore::PgSQL;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 
 use Mail::SpamAssassin::BayesStore::SQL;
 use Mail::SpamAssassin::Logger;
 
-use vars qw( @ISA );
-
-@ISA = qw( Mail::SpamAssassin::BayesStore::SQL );
+our @ISA = qw( Mail::SpamAssassin::BayesStore::SQL );
 
 use constant HAS_DBI => eval { require DBI; };
 
@@ -305,7 +306,7 @@ sub set_running_expire_tok {
   unless (defined($rows)) {
     dbg("bayes: set_running_expire_tok: SQL error: ".$self->{_dbh}->errstr());
     $self->{_dbh}->rollback();
-    return undef;
+    return;
   }
 
   $self->{_dbh}->commit();
@@ -348,7 +349,7 @@ sub remove_running_expire_tok {
 public instance (Integer, Integer, Integer) tok_get (String $token)
 
 Description:
-This method retrieves a specificed token (C<$token>) from the database
+This method retrieves a specified token (C<$token>) from the database
 and returns it's spam_count, ham_count and last access time.
 
 =cut
@@ -397,7 +398,7 @@ public instance (\@) tok_get (@ $tokens)
 
 Description:
 This method retrieves the specified tokens (C<$tokens>) from storage and returns
-an array ref of arrays spam count, ham acount and last access time.
+an array ref of arrays spam count, ham count and last access time.
 
 =cut
 
@@ -587,15 +588,17 @@ sub tok_touch {
   }
 
   my $rows = $sth->rows;
+  # returns the number of rows affected by the last row affecting command,
+  # or -1 if the number of rows is not known or not available
 
-  unless (defined($rows)) {
+  unless (defined($rows) && $rows >= 0) {
     dbg("bayes: tok_touch: SQL error: ".$self->{_dbh}->errstr());
     $self->{_dbh}->rollback();
     return 0;
   }
 
   # if we didn't update a row then no need to update newest_token_age
-  if ($rows eq '0E0') {
+  if (!$rows) {
     $self->{_dbh}->commit();
     return 1;
   }
@@ -638,6 +641,76 @@ do in tok_get_all)
 
 =cut
 
+# tok_touch_all as proposed in bug 6444, executes one update for each token;
+# might run faster with some (older?) versions of PostgreSQL
+# (to switch: rename this one to tok_touch_all and stash the other one away)
+#
+sub tok_touch_all_6444 {
+  my ($self, $tokens, $atime) = @_;
+
+  return 0 unless (defined($self->{_dbh}));
+
+  return 1 unless (scalar(@{$tokens}));
+
+  my $sql =
+    "UPDATE bayes_token SET atime=? WHERE id=? AND token=? AND atime < ?";
+
+  my $sth = $self->{_dbh}->prepare_cached($sql);
+  unless (defined($sth)) {
+    dbg("bayes: tok_touch_all: SQL error: ".$self->{_dbh}->errstr());
+    return 0;
+  }
+
+  my $bytea_type = { pg_type => DBD::Pg::PG_BYTEA };
+  $sth->bind_param(1, $atime);
+  $sth->bind_param(2, $self->{_userid});
+# $sth->bind_param(3, $token, $bytea_type);  # later
+  $sth->bind_param(4, $atime);
+
+  $self->{_dbh}->begin_work();
+
+  my $n_updates = 0;
+  foreach my $token (@$tokens) {
+    $sth->bind_param(3, $token, $bytea_type);
+    my $rc = $sth->execute();
+    unless ($rc) {
+      dbg("bayes: tok_touch_all: SQL error: ".$self->{_dbh}->errstr());
+      $self->{_dbh}->rollback();
+      return 0;
+    }
+    my $rows = $sth->rows;
+    unless (defined($rows) && $rows >= 0) {
+      dbg("bayes: tok_touch_all: SQL error: ".$self->{_dbh}->errstr());
+      $self->{_dbh}->rollback();
+      return 0;
+    }
+    if ($rows > 0) { $n_updates++ }
+  }
+
+  # if we didn't update a row then no need to update newest_token_age
+  if ($n_updates) {  # update newest_token_age
+    # need to check newest_token_age
+    # no need to check oldest_token_age since we would only update if the
+    # atime was newer than what is in the database
+    my $sql_upd_age = "UPDATE bayes_vars SET newest_token_age = ?".
+                      " WHERE id = ? AND newest_token_age < ?";
+    my $rows = $self->{_dbh}->do($sql_upd_age,
+                                 undef, $atime, $self->{_userid}, $atime);
+    unless (defined($rows) && $rows >= 0) {
+      dbg("bayes: tok_touch_all: SQL error: ".$self->{_dbh}->errstr());
+      $self->{_dbh}->rollback();
+      return 0;
+    }
+  }
+  $self->{_dbh}->commit();
+  dbg("bayes: tok_touch_all: updated %d tokens", $n_updates);
+  return 1;
+}
+
+# original tok_touch_all (not the one proposed in bug 6444),
+# executes one update for all token using an IN operator;
+# seems to run faster with PostgreSQL 8.3.14 than the alternative 
+#
 sub tok_touch_all {
   my ($self, $tokens, $atime) = @_;
 
@@ -686,15 +759,17 @@ sub tok_touch_all {
   }
 
   my $rows = $sth->rows;
+  # returns the number of rows affected by the last row affecting command,
+  # or -1 if the number of rows is not known or not available
 
-  unless (defined($rows)) {
+  unless (defined($rows) && $rows >= 0) {
     dbg("bayes: tok_touch_all: SQL error: ".$self->{_dbh}->errstr());
     $self->{_dbh}->rollback();
     return 0;
   }
 
   # if we didn't update a row then no need to update newest_token_age
-  if ($rows eq '0E0') {
+  if (!$rows) {
     $self->{_dbh}->commit();
     return 1;
   }
@@ -720,45 +795,12 @@ sub tok_touch_all {
   return 1;
 }
 
-
-sub tok_touch_allold {
-  my ($self, $tokens, $atime) = @_;
-
-  return 0 unless (defined($self->{_dbh}));
-
-  return 1 unless (scalar(@{$tokens}));
-
-  my $tokenarray = join(",", map { '"' . _quote_bytea($_) . '"' } sort @{$tokens});
-
-  my $sth = $self->{_dbh}->prepare("select touch_tokens($self->{_userid}, $self->{_esc_prefix}'{$tokenarray}', $atime)");
-
-  unless (defined($sth)) {
-    dbg("bayes: tok_touch_all: SQL error: ".$self->{_dbh}->errstr());
-    $self->{_dbh}->rollback();
-    return 0;
-  }
-
-  my $rc = $sth->execute();
-
-  unless ($rc) {
-    dbg("bayes: tok_touch_all: SQL error: ".$self->{_dbh}->errstr());
-    $self->{_dbh}->rollback();
-    return 0;
-  }
-
-  $sth->finish();
-
-  $self->{_dbh}->commit();
-
-  return 1;
-}
-
 =head2 cleanup
 
 public instance (Boolean) cleanup ()
 
 Description:
-This method perfoms any cleanup necessary before moving onto the next
+This method performs any cleanup necessary before moving onto the next
 operation.
 
 =cut
@@ -891,6 +933,10 @@ sub _connect_db {
     dbg("bayes: database connection established");
   }
 
+  if ( $dbh->{pg_server_version} >= 80300 ) {
+    $dbh->do('SET synchronous_commit=off')
+  }
+
   if ( $dbh->{pg_server_version} >= 80100 ) {
     $self->{_esc_prefix} = 'E';
   }
